/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "GetBranchLog.h"
#include "BasicIOParser.h"
#include "Settings.h"

#include <QFont>
#include <QBrush>

// FIXME: make this configurable eventually
const int GetBranchLog::PortionSize = 50;

GetBranchLog::GetBranchLog(QObject * parent, const DatabaseFile & db)
    : QAbstractItemModel(parent), AutomateCommand(0), databaseFile(db)
{}

GetBranchLog::~GetBranchLog()
{}

void GetBranchLog::readAll(const QString & branch)
{
    readMore(branch, -1);
}

void GetBranchLog::readMore(const QString & branch, int count)
{
    mergeRevsBrush = QBrush(QColor(
       Settings::getString("ChangesetBrowserMergeRevisions")
    ));
    outOfBranchRevsBrush = QBrush(QColor(
       Settings::getString("ChangesetBrowserOutOfBranchRevisions")
    ));

    revsRead = 0;
    revsToRead = count;

    if (!branch.isEmpty() && branch != currentBranch)
    {
        AutomateCommand::abortThreads();
        currentBranch = branch;

        if (!nextRevs.contains(currentBranch))
        {
            MonotoneTaskPtr task(new MonotoneTask(QStringList() << "select" << "h:" + branch));
            AutomateCommand::enqueueDatabaseTask(databaseFile, task);
        }
        else
        {
            readNext();
        }

        reset();
        return;
    }

    I(!currentBranch.isEmpty());
    readNext();
}

void GetBranchLog::readNext()
{
    bool canReadNext =
        nextRevs[currentBranch].size() > 0 &&
        (revsToRead == -1 || revsRead < revsToRead);
    if (!canReadNext)
    {
        emit readingStopped();
        return;
    }

    QString next = nextRevs[currentBranch].takeLast();
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "parents" << next));
    AutomateCommand::enqueueDatabaseTask(databaseFile, task);
}

void GetBranchLog::stopReading()
{
    revsToRead = 0;
}

void GetBranchLog::processTaskResult(const MonotoneTaskPtr & task)
{
    if (task->getReturnCode() != 0)
    {
        C(QString("Command returned with a non-zero return code (%1)")
            .arg(task->getDecodedOutput()));
        return;
    }

    QString current = task->getArguments().at(0);
    QString output = task->getDecodedOutput();

    if (current == "toposort")
    {
        QStringList revs = output.split("\n", QString::SkipEmptyParts);
        I(nextRevs[currentBranch].size() == revs.size());
        nextRevs[currentBranch] = revs;

        readNext();
    }
    else
    if (current == "select" || current == "parents")
    {
        // ensure that we never start with a 'select h:' call and have
        // already revisions for the current branch
        I(current == "parents" || revisionMap[currentBranch].size() == 0);

        QStringList revs = output.split("\n", QString::SkipEmptyParts);

        if (current == "parents" && revs.size() > 1)
        {
            mergeRevs.insert(task->getArguments().at(1));
        }

        bool newRevsArrived = false;
        foreach (QString rev, revs)
        {
            // we found a parent which we already recorded
            if (revisionMap[currentBranch].contains(rev))
                continue;

            newRevsArrived = true;
            int newRow = revisionMap[currentBranch].size();
            beginInsertRows(QModelIndex(), newRow, newRow);

            revisionMap[currentBranch].append(rev);
            revsRead++;
            nextRevs[currentBranch].append(rev);

            if (!certMap.contains(rev))
            {
                MonotoneTaskPtr task(new MonotoneTask(QStringList() << "certs" << rev));
                AutomateCommand::enqueueDatabaseTask(databaseFile, task);
            }

            endInsertRows();
        }

        if (newRevsArrived)
        {
            QStringList args;
            args << "toposort";
            args += nextRevs[currentBranch];

            MonotoneTaskPtr task(new MonotoneTask(args));
            AutomateCommand::enqueueDatabaseTask(databaseFile, task);
            return;
        }

        readNext();
    }
    else
    if (current == "certs")
    {
        QString queriedRev = task->getArguments().at(1);
        int row = revisionMap[currentBranch].indexOf(queriedRev);
        I(row >= 0);

        BasicIOParser parser(output);
        I(parser.parse());

        CertList certs;
        certs.revision = queriedRev;
        certs.fill(parser.getStanzas());
        certMap.insert(queriedRev, certs);

        emit dataChanged(index(row, 0, QModelIndex()), index(row, 3, QModelIndex()));
    }
}

int GetBranchLog::columnCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    return 4;
}

QVariant GetBranchLog::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
    {
        return QVariant();
    }

    if (role == Qt::FontRole && index.column() == 3)
    {
        QFont font;
        font.setStyleHint(QFont::Courier);
        font.setFamily("Courier");
        return QVariant(font);
    }

    int row = index.row();
    int col = index.column();

    if (!revisionMap.contains(currentBranch) ||
        row >= revisionMap[currentBranch].count())
    {
        return QVariant();
    }

    QString rev = revisionMap[currentBranch].at(row);

    if (role == Qt::UserRole)
    {
        if (certMap.contains(rev))
        {
            QList<Cert> changeLogCerts = certMap[rev].findByName("changelog");
            QStringList changeLogs;
            foreach (Cert cert, changeLogCerts)
            {
                changeLogs.push_back(cert.htmlValue());
            }

            return QVariant(changeLogs.join("\n<hr>\n"));
        }
    }

    if (role == Qt::ForegroundRole)
    {
        if (certMap.contains(rev))
        {
            QStringList branches = certMap[rev].findByNameValues("branch");

            if (!branches.contains(currentBranch))
            {
                return QVariant(outOfBranchRevsBrush);
            }
        }

        if (mergeRevs.contains(rev))
        {
            return QVariant(mergeRevsBrush);
        }
    }

    if (role == Qt::ToolTipRole)
    {
        if (certMap.contains(rev))
        {
            QStringList certValues;
            foreach (const Cert & cert, certMap[rev].certs)
            {
                if (cert.name == "changelog")
                    continue;

                QString value = cert.value;
                if (cert.name == "date")
                {
                    value = cert.dateTimeValue().toString(
                        Qt::DefaultLocaleShortDate
                    );
                }

                certValues.append(QString("%1: %2").arg(cert.name).arg(value));
            }
            return QVariant(certValues.join("\n"));
        }
    }

    if (role == Qt::DisplayRole)
    {

        if (!certMap.contains(rev))
            return QVariant();

        if (col == 0)
        {
            QList<Cert> dateCerts = certMap[rev].findByName("date");
            QStringList dates;
            foreach (Cert cert, dateCerts)
            {
                dates.push_back(
                    cert.dateTimeValue().toString(
                        Qt::DefaultLocaleShortDate
                    )
                );
            }
            return QVariant(dates.join(", "));
        }
        else if (col == 1)
        {
            QStringList authors = certMap[rev].findByNameValues("author");
            return QVariant(authors.join(", "));
        }
        else if (col == 2)
        {
            QList<Cert> changelogCerts = certMap[rev].findByName("changelog");
            QStringList changelogs;
            foreach (Cert cert, changelogCerts)
            {
                QString plain = cert.plainValue();
                if (plain.size() > 50)
                    plain = plain.left(47).append("...");
                changelogs.push_back(plain);
            }
            return QVariant(changelogs.join(", "));
        }
        else if (col == 3)
        {
            return QVariant(rev);
        }
    }

    return QVariant();
}

Qt::ItemFlags GetBranchLog::flags(const QModelIndex & index) const
{
    if (index.isValid())
    {
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }
    return 0;
}

QVariant GetBranchLog::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    {
        switch(section)
        {
        case(0):
            return QVariant(tr("Date"));
        case(1):
            return QVariant(tr("Author"));
        case(2):
            return QVariant(tr("Changelog"));
        case(3):
            return QVariant(tr("Revision ID"));
        }
    }
    return QVariant();
}

int GetBranchLog::rowCount(const QModelIndex & parent) const
{
    Q_UNUSED(parent);
    if (!revisionMap.contains(currentBranch))
        return 0;
    return revisionMap[currentBranch].count();
}

QModelIndex GetBranchLog::index(int row, int column, const QModelIndex & parent) const
{
    if (!hasIndex(row, column, parent))
    {
        return QModelIndex();
    }
    return createIndex(row, column, 0);
}

QModelIndex GetBranchLog::parent(const QModelIndex & index) const
{
    Q_UNUSED(index);
    return QModelIndex();
}

