/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneManager.h"
#include "MonotoneProcess.h"
#include "MonotoneUtil.h"
#include "vocab.h"
#include "BasicIOParser.h"

#include <QMetaType>
#include <QMessageBox>

const QString MonotoneManager::MinInterfaceVersion = MIN_MTN_INT_VERSION;
const QString MonotoneManager::MaxInterfaceVersion = MAX_MTN_INT_VERSION;

MonotoneManager::MonotoneManager() : threadNumber(0) {}

MonotoneManager::~MonotoneManager()
{
    QMapIterator<int, MonotoneThreadPtr> it(threadMap);
    while (it.hasNext())
    {
        it.next();
        MonotoneThreadPtr thread = it.value();
        thread->abort();
        thread->wait();
    }
    threadMap.clear();
}

bool MonotoneManager::setMonotoneBinaryPath(const QString & mtnBinary,
    bool relaxedVersionCheck, QString & pathVersion)
{
    pathVersion = getInterfaceVersion(mtnBinary);

    // check if we match the minimum interface version
    if (versionCompare(pathVersion, MinInterfaceVersion) < 0)
    {
        return false;
    }

    // check if we should check and match also the maximum interface version
    if (!relaxedVersionCheck && versionCompare(pathVersion, MaxInterfaceVersion) > 0)
    {
        return false;
    }

    mtnPath = mtnBinary;
    return true;
}

MonotoneThreadPtr MonotoneManager::getThreadForWorkspace(const WorkspacePath & workspace)
{
    // ensure that its normalized
    WorkspacePath normalizedWorkspace = normalizeWorkspacePath(workspace);
    DatabaseFile databaseFile = getDatabaseFilePath(normalizedWorkspace);
    return getThread(databaseFile, normalizedWorkspace);
}

MonotoneThreadPtr MonotoneManager::getThreadForDatabase(const DatabaseFile & database)
{
    if (database.isEmpty())
    {
        throw GuitoneException(tr("Database file path is empty"));
    }

    QFile dbfile(database);
    if (!dbfile.open(QIODevice::ReadOnly))
    {
        throw GuitoneException(
           tr("Could not open database for reading: %1").arg(dbfile.errorString())
        );
    }

    if (dbfile.read(15) != QByteArray("SQLite format 3"))
    {
        dbfile.close();
        throw GuitoneException(tr("Database file not an SQLite 3 database"));
    }

    if (!dbfile.seek(60) || dbfile.read(4) != QByteArray("_MTN"))
    {
        dbfile.close();
        throw GuitoneException(tr("SQLite database is not a monotone "
                                  "database or a database of an older, "
                                  "unsupported version of monotone"));
    }
    dbfile.close();
    return getThread(database, QString());
}

MonotoneThreadPtr MonotoneManager::getThread(const DatabaseFile & database, const WorkspacePath & workspace)
{
    QMutexLocker locker(&lock);

    if (mtnPath.isEmpty())
        throw GuitoneException(tr("Monotone binary path not set"));

    // since we cannot set the workspace directory after we've
    // started the process, we need to ensure that each workspace runs with
    // its own process
    QList<int> threadIDs = identMap.keys(database + "|" + workspace);

    QList<int> availableThreads;
    foreach (int threadID, threadIDs)
    {
        if (stoppingThreads.contains(threadID))
            continue;
        availableThreads.append(threadID);
    }

    if (availableThreads.size() == 0)
    {
        int newThreadID = ++threadNumber;

        MonotoneThreadPtr thread
            (new MonotoneThread(newThreadID, mtnPath, database, workspace));

        threadMap.insert(newThreadID, thread);
        identMap.insert(newThreadID, database + "|");
        if (!workspace.isEmpty())
        {
            identMap.insert(newThreadID, database + "|" + workspace);
        }
        availableThreads.append(newThreadID);

        connect(
            thread.data(), SIGNAL(stopped(int)),
            this, SLOT(stopped(int))
        );

        connect(
            thread.data(), SIGNAL(aborted(int, QProcess::ProcessError, const QString &)),
            this, SLOT(aborted(int, QProcess::ProcessError, const QString &))
        );

        thread->start();
    }

    // TODO: we may want to add support for multiple threads for one
    // and the same database here in the future... then we should probably
    // also make a smarter choice of available threads here instead of just
    // taking the first!
    return threadMap.value(availableThreads.value(0));
}

void MonotoneManager::stopThread(int threadID)
{
    QMutexLocker locker(&lock);

    if (!threadMap.contains(threadID))
    {
        throw GuitoneException(tr("No such thread: %1").arg(threadID));
    }

    L(QString("stopping thread %1").arg(threadID));
    MonotoneThreadPtr thread = threadMap.value(threadID);
    thread->abort();

    stoppingThreads.insert(threadID);
}

void MonotoneManager::stopWorkspaceThreads(const WorkspacePath & workspace)
{
    WorkspacePath normalizedWorkspace = normalizeWorkspacePath(workspace);
    DatabaseFile databaseFile = getDatabaseFilePath(normalizedWorkspace);
    QList<int> threadIDs = identMap.keys(databaseFile + "|" + normalizedWorkspace);

    L(QString("stopping %1 threads for workspace %2")
        .arg(threadIDs.size()).arg(normalizedWorkspace));

    foreach (int threadID, threadIDs)
    {
        stopThread(threadID);
    }
}

WorkspacePath MonotoneManager::normalizeWorkspacePath(const WorkspacePath & workspace)
{
    QDir tempDir(workspace);
    if (!tempDir.exists())
    {
        throw GuitoneException(tr("Directory %1 does not exist").arg(workspace));
    }

    bool found = false;
    WorkspacePath normalizedWorkspace;
    do
    {
        if (tempDir.cd("_MTN"))
        {
            tempDir.cdUp();
            normalizedWorkspace = tempDir.absolutePath();
            found = true;
            break;
        }
    }
    while (!tempDir.isRoot() && tempDir.cdUp());

    if (!found)
    {
        throw GuitoneException(tr("could not find _MTN directory"));
    }

    return normalizedWorkspace;
}

// we assume that the workspace was already normalized here
DatabaseFile MonotoneManager::getDatabaseFilePath(const WorkspacePath & workspace)
{
    // now check again if we know it
    if (workspaceMap.contains(workspace))
    {
        return workspaceMap.value(workspace);
    }

    QFile optionsFile(workspace + "/_MTN/options");
    if (!optionsFile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        throw GuitoneException(tr("could not open _MTN/options for reading"));
    }

    QByteArray contents = optionsFile.readAll();
    optionsFile.close();

    if (contents.size() == 0)
    {
        throw GuitoneException(tr("file _MTN/options is empty"));
    }

    BasicIOParser parser(QString::fromUtf8(contents));
    if (!parser.parse())
    {
        throw GuitoneException(tr("could not parse basic_io from _MTN/options"));
    }
    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1);
    Stanza st = stanzas.at(0);

    DatabaseFile databaseFilePath;
    foreach (const StanzaEntry & entry, st)
    {
        if (entry.sym == "database")
        {
            I(entry.vals.size() == 1);
            databaseFilePath = entry.vals.at(0);
            break;
        }
    }

    if (databaseFilePath.isEmpty())
    {
        throw GuitoneException(tr("could not find database for workspace"));
    }

    // remember what we've just found for later requests
    workspaceMap.insert(workspace, databaseFilePath);
    return databaseFilePath;
}

void MonotoneManager::removeThread(int threadID)
{
    QMutexLocker locker(&lock);

    if (!threadMap.contains(threadID))
    {
        L(QString("thread %1 was already removed").arg(threadID));
        return;
    }

    MonotoneThreadPtr thread = threadMap.value(threadID);

    disconnect(
        thread.data(), SIGNAL(stopped(int)),
        this, SLOT(stopped(int))
    );

    disconnect(
        thread.data(), SIGNAL(aborted(int, QProcess::ProcessError, const QString &)),
        this, SLOT(aborted(int, QProcess::ProcessError, const QString &))
    );

    threadMap.remove(threadID);
    identMap.remove(threadID);
    stoppingThreads.remove(threadID);

    // wait until the thread has finished execution
    thread->wait();
}

void MonotoneManager::stopped(int threadID)
{
    removeThread(threadID);
    L(QString("thread %1 stopped").arg(threadID));
}

void MonotoneManager::aborted(int threadID, QProcess::ProcessError error, const QString & message)
{
    I(threadMap.contains(threadID));

    MonotoneThreadPtr thread = threadMap.value(threadID);

    QString databaseFile  = thread->getDatabaseFilePath();
    QString workspacePath = thread->getWorkspacePath();

    removeThread(threadID);

    QString mtnError = MonotoneUtil::stripMtnPrefix(message);
    QString processErrorTranslated;

    switch (error)
    {
        case QProcess::FailedToStart:
            processErrorTranslated = tr("failed to start");
            break;
        case QProcess::Crashed:
            processErrorTranslated = tr("crashed");
            break;
        case QProcess::Timedout:
            processErrorTranslated = tr("timed out waiting for data");
            break;
        case QProcess::WriteError:
            processErrorTranslated = tr("failed to read");
            break;
        case QProcess::ReadError:
            processErrorTranslated = tr("failed to write");
            break;
        case QProcess::UnknownError:
            processErrorTranslated = tr("unknown error");
            break;
        default: I(false);
    }

    // log something reasonable
    C(QString("thread %1 died (%2) - stderr was %3")
        .arg(threadID).arg(processErrorTranslated).arg(mtnError));

    // display something understandable to the user
    QString userMessage(tr(
        "The monotone process for '%1' died (%2). "
        "If you think this is a bug in guitone, "
        "please report it to the author!"
    ).arg(workspacePath.isEmpty() ? databaseFile : workspacePath)
     .arg(processErrorTranslated)
    );

    if (!mtnError.isEmpty())
    {
        userMessage.append(tr("\n\nLast error output was:\n%1").arg(mtnError));
    }

    emit threadAborted(userMessage);
}

QString MonotoneManager::getInterfaceVersion(const QString & mtnBinary)
{
    MonotoneProcess proc;
    proc.setMtnBinaryPath(mtnBinary);
    proc.start(QStringList() << "automate" << "interface_version");
    proc.waitForFinished();

    QString output = proc.getBufferedOutput();
    if (!proc.successful())
    {
        C(QString("Couldn't execute `automate interface_version`: %1")
          .arg(output));
        return QString();
    }

    QRegExp regex("^(\\d+(?:\\.\\d+))");
    if (regex.indexIn(output) == -1)
    {
        D(QString("couldn't parse output: %1").arg(output));
        return false;
    }

    return regex.cap(1);
}

/*!
    Returns
    -1 if right is greater
    0 if both are equal
    1 if left is greater

    If a string addition ("flag") is supplied to a particular argument,
    a version without such a flag is considered newer than a version with
    such a flag, i.e. 0.9dev < 0.9, but 0.9dev > 0.8. Also, if both
    versions contain a flag on the same position, these flags are string-
    compared, i.e. 0.9 > 0.9rc2 > 0.9rc1 > 0.9dev > 0.8

    \sa checkInterfaceVersion()
*/
int MonotoneManager::versionCompare(const QString & left, const QString & right)
{
    QStringList leftParts = left.split(".");
    QStringList rightParts = right.split(".");

    int leftCount = leftParts.size();
    int rightCount = rightParts.size();
    int maxCount = leftCount > rightCount ? leftCount : rightCount;

    QRegExp version("(\\d+)(.*)");

    for (int i=0, j=maxCount; i<j; i++)
    {
        unsigned int l = 0;
        QString lFlag;
        if (i < leftCount && version.exactMatch(leftParts.at(i)))
        {
            l = version.cap(1).toUInt();
            lFlag = version.cap(2);
        }

        unsigned int r = 0;
        QString rFlag;
        if (i < rightCount && version.exactMatch(rightParts.at(i)))
        {
            r = version.cap(i).toUInt();
            rFlag = version.cap(2);
        }

        if (l == r && lFlag == rFlag)
            continue;

        if (l > r || (l == r && (lFlag.isEmpty() || lFlag > rFlag)))
            return 1;

        if (l < r || (l == r && (rFlag.isEmpty() || rFlag > lFlag)))
            return -1;

        I(false);
    }

    return 0;
}
