/***************************************************************************
 *   Copyright (C) 2007 by Thomas Keller                                   *
 *   me@thomaskeller.biz                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "MonotoneUtil.h"
#include "GuitoneCore.h"
#include "BasicIOParser.h"

void MonotoneUtil::taskFinished(const MonotoneTaskPtr & task)
{
    if (task->getCommandNumber() != currentCommandNumber)
        return;

    emit finished();
}

void MonotoneUtil::taskAborted(const MonotoneTaskPtr & task)
{
    if (task->getCommandNumber() != currentCommandNumber)
        return;

    C(QString("Task %1 (thread %2) aborted (%3)")
        .arg(task->getCommandNumber())
        .arg(task->getThreadNumber())
        .arg(QString::fromUtf8(task->getEncodedInput()))
    );

    emit finished();
}

void MonotoneUtil::run(const MonotoneThreadPtr & thread, const MonotoneTaskPtr & task)
{
    connect(
        thread.data(), SIGNAL(taskFinished(const MonotoneTaskPtr &)),
        this, SLOT(taskFinished(const MonotoneTaskPtr &))
    );

    connect(
        thread.data(), SIGNAL(taskAborted(const MonotoneTaskPtr &)),
        this, SLOT(taskAborted(const MonotoneTaskPtr &))
    );

    currentCommandNumber = thread->enqueueTask(task);

    QEventLoop * ev = new QEventLoop(this);
    connect(this, SIGNAL(finished()), ev, SLOT(quit()));
    ev->exec();

    delete ev;

    disconnect(
        thread.data(), SIGNAL(taskFinished(const MonotoneTaskPtr &)),
        this, SLOT(taskFinished(const MonotoneTaskPtr &))
    );

    disconnect(
        thread.data(), SIGNAL(taskAborted(const MonotoneTaskPtr &)),
        this, SLOT(taskAborted(const MonotoneTaskPtr &))
    );
}

void MonotoneUtil::runSynchronousWorkspaceTask(const WorkspacePath & workspace, const MonotoneTaskPtr & task)
{
    MonotoneUtil * instance = new MonotoneUtil();
    instance->run(APP->manager()->getThreadForWorkspace(workspace), task);
    delete instance;
}

void MonotoneUtil::runSynchronousDatabaseTask(const DatabaseFile & database, const MonotoneTaskPtr & task)
{
    MonotoneUtil * instance = new MonotoneUtil();
    instance->run(APP->manager()->getThreadForDatabase(database), task);
    delete instance;
}

DatabaseFile MonotoneUtil::getDatabaseFile(const WorkspacePath & workspace)
{
    return APP->manager()->getDatabaseFilePath(workspace);
}

QString MonotoneUtil::getBaseWorkspaceRevision(const WorkspacePath & workspace)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_base_revision_id"));
    runSynchronousWorkspaceTask(workspace, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QString();
    }

    QString data = task->getDecodedOutput();
    if (task->getReturnCode() > 0)
    {
        C(QString("Could not execute get_base_revision_id: %1").arg(data));
        return QString();
    }

    data.chop(1);

    // we cannot handle workspaces with multiple parents yet
    I(data.indexOf('\n') == -1);

    return data;
}

QString MonotoneUtil::getOption(const WorkspacePath & workspace, const QString & opt)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_option" << opt));
    runSynchronousWorkspaceTask(workspace, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QString();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't retrieve option %1: %2")
            .arg(opt).arg(task->getLast(MonotoneTask::Error)));
        return QString();
    }

    QString data = task->getDecodedOutput();
    // remove the trailing newline
    data.chop(1);

    return data;
}

QString MonotoneUtil::getBranchName(const WorkspacePath & workspace, const QString & defaultBranch)
{
    QString branchName = getOption(workspace, "branch");

    if (branchName.size() == 0)
    {
        branchName = defaultBranch;
    }
    return branchName;
}

QString MonotoneUtil::getBranchNameShort(const WorkspacePath & workspace)
{
    // FIXME: currently there is only support for the domain branch scheme
    // tld.domain.project.branch
    QString branchName = getBranchName(workspace);
    QStringList parts = branchName.split('.');
    if (parts.size() == 1) return branchName;

    int maxSize = 30;
    int minNeededChars = (parts.size() * 2) - 1;

    QString shortBranchName;
    for (int i=parts.size()-1; i>=0 ; i--)
    {
        QString part = parts.at(i);
        if ((part.size() + shortBranchName.size() + minNeededChars) > maxSize)
        {
            shortBranchName.prepend(part.at(0));
        }
        else
        {
            shortBranchName.prepend(part);
        }

        if (i != 0) shortBranchName.prepend(".");
        minNeededChars -= 2;
    }

    return shortBranchName;
}

QStringList MonotoneUtil::resolveSelector(const DatabaseFile & db, const QString & selector)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "select" << selector));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QStringList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't resolve selector %1: %2")
            .arg(selector).arg(task->getLast(MonotoneTask::Error)));
        return QStringList();
    }

    QString data = task->getDecodedOutput();
    QStringList revList = data.split('\n', QString::SkipEmptyParts);
    return revList;
}

CertList MonotoneUtil::getCerts(const DatabaseFile & db, const QString & revision)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "certs" << revision));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return CertList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query revision certs for %1: %2")
            .arg(revision).arg(task->getLast(MonotoneTask::Error)));
        return CertList();
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    CertList certs;
    certs.revision = revision;
    certs.fill(parser.getStanzas());
    return certs;
}

FileEntryList MonotoneUtil::getRevisionManifest(const DatabaseFile & db, const QString & revision)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_manifest_of" << revision));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return FileEntryList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query manifest entries for %1: %2")
            .arg(revision).arg(task->getLast(MonotoneTask::Error)));
        return FileEntryList();
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    // FIXME: this should go into a separate function / class
    FileEntryList entries;
    StanzaList stanzas = parser.getStanzas();
    foreach (const Stanza & st, stanzas)
    {
        FileEntry entry;
        bool is_item = false;

        foreach (const StanzaEntry & en, st)
        {
            if (en.sym == "format_version")
            {
                I(en.vals.size() == 1 && en.vals.at(0) == "1");
                break;
            }
            else
            if (en.sym == "dir" || en.sym == "file")
            {
                is_item = true;
                entry.path = en.vals.at(0);
                entry.is_dir = en.sym == "dir";
            }
            else
            if (en.sym == "content")
            {
                I(is_item);
                entry.fileid = en.hash;
            }
            else
            if (en.sym == "attr")
            {
                I(is_item);
                I(en.vals.size() == 2);
                entry.attrs.insert(en.vals.at(0), en.vals.at(1));
            }
            else
                I(false);
        }
        if (is_item) entries.append(entry);
    }
    return entries;
}

QMap<QString, QString> MonotoneUtil::getPublicKeyMap(const DatabaseFile & db)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "keys"));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QMap<QString, QString>();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query keys: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return QMap<QString, QString>();
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    QMap<QString, QString> keys;
    StanzaList stanzas = parser.getStanzas();
    foreach (const Stanza & st, stanzas)
    {
        QString lastHash;
        QString lastLocalName;

        foreach (const StanzaEntry & en, st)
        {
            if (en.sym == "hash")
            {
                lastHash = en.hash;
                continue;
            }

            if (en.sym == "local_name")
            {
                lastLocalName = en.vals.at(0);
                continue;
            }

            if (en.sym == "public_location")
            {
                I(!lastHash.isEmpty() && !lastLocalName.isEmpty());
                keys.insert(lastHash, QObject::tr("%1 (%2...)")
                            .arg(lastLocalName).arg(lastHash.left(8)));
                break;
            }
        }
    }
    return keys;
}

QMap<QString, QString> MonotoneUtil::getPrivateKeyMap(const DatabaseFile & db)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "keys"));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QMap<QString, QString>();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query keys: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return QMap<QString, QString>();
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    QMap<QString, QString> keys;
    StanzaList stanzas = parser.getStanzas();
    foreach (const Stanza & st, stanzas)
    {
        QString lastHash;
        QString lastGivenName;

        foreach (const StanzaEntry & en, st)
        {
            if (en.sym == "hash")
            {
                lastHash = en.hash;
                continue;
            }

            if (en.sym == "given_name")
            {
                lastGivenName = en.vals.at(0);
                continue;
            }

            if (en.sym == "private_location")
            {
                I(!lastHash.isEmpty() && !lastGivenName.isEmpty());
                keys.insert(lastHash, QObject::tr("%1 (%2...)")
                            .arg(lastGivenName).arg(lastHash.left(8)));
                break;
            }
        }
    }
    return keys;
}

QString MonotoneUtil::getFileId(const DatabaseFile & db, const QString & path)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "identify" << path));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QString();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't identify path: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return QString();
    }

    QString data = task->getDecodedOutput();
    data.chop(1);
    return data;
}

QStringList MonotoneUtil::getRoots(const DatabaseFile & db, bool & ok)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "roots"));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        ok = false;
        return QStringList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query roots: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        ok = false;
        return QStringList();
    }

    ok = true;
    QString data = task->getDecodedOutput();
    return data.split('\n', QString::SkipEmptyParts);
}

QStringList MonotoneUtil::getCommonAncestors(const DatabaseFile & db, const QStringList & revs)
{
    QStringList args;
    args << "common_ancestors";
    args += revs;

    MonotoneTaskPtr task(new MonotoneTask(args));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QStringList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query common ancestors: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return QStringList();
    }

    QString data = task->getDecodedOutput();
    return data.split('\n', QString::SkipEmptyParts);
}

QStringList MonotoneUtil::getPreviousContentMarks(const DatabaseFile & db, const QString & file, const QString & startRev)
{
    QStringList args;
    args << "get_content_changed" << startRev << file;
    MonotoneTaskPtr task(new MonotoneTask(args));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QStringList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't query content marks for '%1' (%2)")
            .arg(file).arg(task->getLast(MonotoneTask::Error)));
        return QStringList();
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());

    QStringList revs;
    foreach (const Stanza & st, parser.getStanzas())
    {
        I(st.size() == 1);
        revs.push_back(st.at(0).hash);
    }
    return revs;
}

QStringList MonotoneUtil::topsortRevisions(const DatabaseFile & db, const QStringList & revs)
{
    QStringList args;
    args << "toposort";
    args += revs;
    MonotoneTaskPtr task(new MonotoneTask(args));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QStringList();
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't toposort revisions: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        // return the list unchanged
        return revs;
    }

    QString data = task->getDecodedOutput();
    return data.split("\n", QString::SkipEmptyParts);
}

bool MonotoneUtil::getAttribute(const WorkspacePath & ws, const QString & path,
                  const QString & attrname, QPair<QString, QString> & attrval)
{
    MonotoneTaskPtr task(new MonotoneTask(QStringList() << "get_attributes" << path));
    runSynchronousWorkspaceTask(ws, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return false;
    }

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't run get_attributes: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return false;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();

    foreach (const Stanza & st, stanzas)
    {
        if (st.at(0).sym != "attr") continue;
        if (st.at(0).vals.at(0) == attrname)
        {
            attrval.first = st.at(0).vals.at(1);
            attrval.second = st.at(1).vals.at(0);
            return true;
        }
    }

    return false;
}

bool MonotoneUtil::getAttribute(const DatabaseFile & db, const QString & revision,
                  const QString & path, const QString & attrname, QString & attrval)
{
    FileEntryList entryList = getRevisionManifest(db, revision);

    foreach (const FileEntry & en, entryList)
    {
        if (en.path != path) continue;
        if (en.attrs.contains(attrname))
        {
            attrval = en.attrs.value(attrname);
            return true;
        }
    }

    return false;
}

QMap<QString, QString> MonotoneUtil::getDatabaseVariables(const DatabaseFile & db,
                                                          const QString & domain)
{
    QStringList args;
    args << "get_db_variables" << domain;
    MonotoneTaskPtr task(new MonotoneTask(args));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QMap<QString, QString>();
    }

    QMap<QString, QString> entries;

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't run get_db_variables: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return entries;
    }

    BasicIOParser parser(task->getDecodedOutput());
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();

    if (stanzas.size() > 0)
    {
        foreach (StanzaEntry st, stanzas.at(0))
        {
            if (st.sym == "domain") continue;
            I(st.sym == "entry" && st.vals.size() == 2);
            entries.insert(st.vals[0], st.vals[1]);
        }
    }

    return entries;
}

QString MonotoneUtil::getCorrespondingPath(const DatabaseFile & db,
                                                          const QString & rev1,
                                                          const QString & filePath,
                                                          const QString & rev2)
{
    QStringList args;
    args << "get_corresponding_path" << rev1 << filePath << rev2;
    MonotoneTaskPtr task(new MonotoneTask(args));
    runSynchronousDatabaseTask(db, task);
    if (!task->isFinished())
    {
        C(QString("task '%1' aborted").arg(QString(task->getEncodedInput())));
        return QString();
    }

    QMap<QString, QString> entries;

    if (task->getReturnCode() > 0)
    {
        C(QString("Couldn't run get_corresponding_path: %1")
            .arg(task->getLast(MonotoneTask::Error)));
        return QString();
    }

    QString out(task->getDecodedOutput());
    if (out.isEmpty())
        return QString();

    BasicIOParser parser(out);
    I(parser.parse());
    StanzaList stanzas = parser.getStanzas();
    I(stanzas.size() == 1 && stanzas.at(0).size() == 1);
    return stanzas.at(0).at(0).vals.at(0);
}

/*!
    Strips mtn: warning: and alike from error strings coming from mtn
    and unwraps them, if this fails we add the original string unwrapped
*/
QString MonotoneUtil::stripMtnPrefix(const QString & input)
{
    QString output;
    QStringList list = input.split(QRegExp("\\n"));

    // FIXME: we should actually use basename(mtnBinaryPath) in case
    // the binary is renamed, but I'm too lazy for this right now...
    // and this would also require to make this method non-static
    QRegExp rx = QRegExp("^(?:mtn\\:[ ])?\\w+\\:[ ]+(.+)");
    rx.setPatternSyntax(QRegExp::RegExp2);
    for (int i=0, j=list.size(); i<j; i++)
    {
        if (rx.indexIn(list[i]) != -1)
        {
            output += rx.cap(1) + " ";
            continue;
        }
        output += list[i] + " ";
    }
    return output.trimmed();
}

