/***************************************************************************
 *   Copyright (C) 2006 by Jean-Louis Fuchs                                 *
 *   ganwell@fangorn.ch                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#include "ChangesetBrowser.h"
#include "Settings.h"

ChangesetBrowser::ChangesetBrowser(QWidget * parent, const DatabaseFile & db)
  : Dialog(parent), databaseFile(db), branchModel(0)
{
    setupUi(this);
    Dialog::init();

    this->setWindowFlags(this->windowFlags() | Qt::WindowMaximizeButtonHint);
    this->setWindowFlags(this->windowFlags() | Qt::WindowMinimizeButtonHint);

    multipleParents->setVisible(false);
    progressWidget->setVisible(false);
    fetchWidget->setVisible(false);

    branchesSplitter->restoreState();
    changesetsSplitter->restoreState();
    changeLogSplitter->restoreState();

    tree = Settings::getBool("ChangesetBrowserTree");
    displayBranchesAsTree->setText(
        !tree ? tr("display branches as tree") : tr("display branches flat")
    );

    initTreeWidget();

    connect(
        branches, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(branchesClicked(const QModelIndex &))
    );

    connect(
        displayBranchesAsTree, SIGNAL(clicked()),
        this, SLOT(toggleTree())
    );

    branchLogModel = new GetBranchLog(this, databaseFile);
    changesets->setModel(branchLogModel);
    changesets->setRootIsDecorated(false);

    connect(
        branchLogModel, SIGNAL(readingStopped()),
        this, SLOT(readingStopped())
    );

    connect(
        this, SIGNAL(dialogHidden()),
        branchLogModel, SLOT(stopReading())
    );

    revisionModel = new GetRevision(this);
    revisionView->setModel(revisionModel);

    connect(
        revisionModel, SIGNAL(revisionRead()),
        this, SLOT(updateParentsList())
    );

    connect(
        revisionParents, SIGNAL(currentIndexChanged(const QString &)),
        revisionModel, SLOT(showChangesAgainstParent(const QString &))
    );

    connect(
        changesets, SIGNAL(clicked(const QModelIndex &)),
        this, SLOT(changesetsClicked(const QModelIndex &))
    );

    connect(
        pushAll, SIGNAL(clicked()),
        this, SLOT(readAll())
    );

    connect(
        pushMore, SIGNAL(clicked()),
        this, SLOT(readMore())
    );

    connect(
        pushStop, SIGNAL(clicked()),
        branchLogModel, SLOT(stopReading())
    );

    connect(
        changesets, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(changesetsDoubleClicked(const QModelIndex &))
    );

    connect(
        changesets, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(changesetsContextMenuRequested(const QModelIndexList &, const QPoint &))
    );

    connect(
        revisionView, SIGNAL(doubleClicked(const QModelIndex &)),
        this, SLOT(revisionViewDoubleClicked(const QModelIndex &))
    );

    connect(
        revisionView, SIGNAL(contextMenuRequested(const QModelIndexList &, const QPoint &)),
        this, SLOT(revisionViewContextMenuRequested(const QModelIndexList &, const QPoint &))
    );
}

ChangesetBrowser::~ChangesetBrowser()
{
    branchesSplitter->saveState();
    changesetsSplitter->saveState();
    changeLogSplitter->saveState();

    delete branchLogModel;
    delete revisionModel;
    if (branchModel) delete branchModel;
    Settings::setBool("ChangesetBrowserTree", tree);
}

void ChangesetBrowser::readAll()
{
    progressWidget->setVisible(true);
    branchLogModel->readAll();
}

void ChangesetBrowser::readMore()
{
    progressWidget->setVisible(true);
    branchLogModel->readMore();
}

void ChangesetBrowser::readingStopped()
{
    progressWidget->setVisible(false);

    // yes, this looks awkward, but there is no other sane
    // way to "reset" the state of a set of exclusive buttons
    pushMore->setAutoExclusive(false);
    pushMore->setChecked(false);
    pushMore->setAutoExclusive(true);

    pushAll->setAutoExclusive(false);
    pushAll->setChecked(false);
    pushAll->setAutoExclusive(true);
}

void ChangesetBrowser::toggleTree()
{
    tree = !tree;
    displayBranchesAsTree->setText(
        !tree ? tr("display branches as tree") : tr("display branches flat")
    );
    initTreeWidget();
}

void ChangesetBrowser::branchesClicked(const QModelIndex & idx)
{
    QString branch = branchModel->data(idx, Qt::ToolTipRole).toString();

    fetchWidget->setVisible(true);
    progressWidget->setVisible(true);
    pushMore->setDown(false);

    branchLogModel->readMore(branch);
}

void ChangesetBrowser::changesetsClicked(const QModelIndex & idx)
{
    QModelIndex revIdx = branchLogModel->index(idx.row(), 3, QModelIndex());
    currentRevision = branchLogModel->data(revIdx, Qt::DisplayRole).toString();
    changeLog->setHtml(branchLogModel->data(revIdx, Qt::UserRole).toString());
    revisionModel->readDatabaseRevision(databaseFile, currentRevision);
    revisionParents->clear();
    multipleParents->setVisible(false);
}

void ChangesetBrowser::initTreeWidget()
{
    if (branchModel)
    {
        disconnect(
            branchModel, SIGNAL(branchesRead(void)),
            this, SLOT(branchesRead(void))
        );
        delete branchModel;
    }

    branchModel = new Branches(this, databaseFile, tree);
    branchModel->readBranches();
    branches->setRootIsDecorated(false);
    branches->setModel(branchModel);

    connect(
        branchModel, SIGNAL(branchesRead(void)),
        this, SLOT(branchesRead(void))
    );
}

void ChangesetBrowser::branchesRead()
{
    branches->setExpanded(branchModel->index(0, 0, QModelIndex()), true);
}

void ChangesetBrowser::changesetsDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;
    QModelIndex revIdx = branchLogModel->index(index.row(), 3, QModelIndex());
    emit revisionManifest(revIdx.data().toString());
}

void ChangesetBrowser::changesetsContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QModelIndex revIdx =
        branchLogModel->index(indexList.at(0).row(), 3, QModelIndex());

    QString rev(revIdx.data().toString());

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QAction * actOpenManifest = menu.addAction(tr("View files in this revision"));
    QAction * actCopyRevisionId = menu.addAction(tr("Copy revision ID"));

    actOpenManifest->setFont(activeFont);

    QAction * act = menu.exec(pos);
    if (act == actOpenManifest)
    {
        emit revisionManifest(rev);
    }
    else
    if (act == actCopyRevisionId)
    {
        QClipboard * clipboard = QApplication::clipboard();
        clipboard->setText(rev);
    }
}

void ChangesetBrowser::updateParentsList()
{
    QStringList parentRevs = revisionModel->getParentRevisions();
    if (parentRevs.size() > 1)
    {
        multipleParents->setVisible(true);
        revisionParents->addItems(revisionModel->getParentRevisions());
    }
}

void ChangesetBrowser::revisionViewDoubleClicked(const QModelIndex & index)
{
    if (!index.isValid()) return;

    QModelIndex label, path;
    if (index.column() == 0)
    {
        label = index;
        path = index.sibling(index.row(), 1);
    }
    else
    if (index.column() == 1)
    {
        label = index.sibling(index.row(), 0);
        path = index;
    }
    else
        I(false);

    // not a patched file
    if (label.data().toString() != QObject::tr("Patched"))
        return;

    emit diffFile(path.data().toString(),
                  revisionModel->getCurrentParentRevision(),
                  currentRevision);
}

void ChangesetBrowser::revisionViewContextMenuRequested(const QModelIndexList & indexList, const QPoint & pos)
{
    if (indexList.size() == 0) return;

    QModelIndex index = indexList.at(0);

    QMenu menu(this);
    QFont activeFont;
    activeFont.setBold(true);

    QModelIndex label, path;
    if (index.column() == 0)
    {
        label = index;
        path = index.sibling(index.row(), 1);
    }
    else
    if (index.column() == 1)
    {
        label = index.sibling(index.row(), 0);
        path = index;
    }
    else
        I(false);


    if (label.data().toString() != QObject::tr("Patched"))
        return;

    QAction * actShowFileDiff = menu.addAction(tr("Show differences"));
    QAction * actAnnotate = menu.addAction(tr("Annotate"));

    actShowFileDiff->setFont(activeFont);

    QAction * act = menu.exec(pos);
    if (act == actShowFileDiff)
    {
         emit diffFile(path.data().toString(),
                       revisionModel->getCurrentParentRevision(),
                       currentRevision);
    }
    else
    if (act == actAnnotate)
    {
        emit annotateFile(currentRevision, path.data().toString());
    }
}

void ChangesetBrowser::closeEvent(QCloseEvent * event)
{
    branchLogModel->stopReading();
    event->accept();
}

