#ifndef VOCAB_H
#define VOCAB_H

//
// global macros and defines
//

// if you use APP, you also have to include GuitoneCore.h
class GuitoneCore;
#define APP reinterpret_cast<GuitoneCore *>(qApp)

#include "DebugLog.h"
#ifdef QT_NO_DEBUG
#define D(msg) void(msg)
#else
#define D(msg) DebugLog::debug(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#endif

#define L(msg) DebugLog::info(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#define W(msg) DebugLog::warn(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))
#define C(msg) DebugLog::critical(QString("%1:%2:%3: %4") \
        .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg)))

// for abort()
#include <stdlib.h>
#define F(msg) \
    do {  DebugLog::fatal(QString("%1:%2:%3: %4") \
          .arg(__FILE__).arg(__FUNCTION__).arg(__LINE__).arg(QString(msg))); \
          abort(); \
    } while (0);

#define I(expr) \
    do { if (!(expr)) { F(QString("invariant \"%1\" violated").arg(#expr)) } } while (0);

//
// type definitions
//

#include <QMap>
#include <QPair>
#include <QList>
#include <QStringList>

// used for BasicIOParser and BasicIOWriter
struct StanzaEntry
{
    QString sym;
    QString hash;
    QStringList vals;

    StanzaEntry() {}

    StanzaEntry(QString s, QString h) : sym(s), hash(h) {}

    StanzaEntry(QString s, QStringList v) : sym(s), vals(v) {}
};

typedef QList <StanzaEntry> Stanza;
typedef QList <Stanza> StanzaList;

// used for manifest entries, if the bool var is true, the entry is a directory
struct FileEntry
{
    QString path;
    bool is_dir;
    QString fileid;
    QMap<QString, QString> attrs;

    FileEntry() {}

    FileEntry(QString p, bool d) : path(p), is_dir(d) {}

    FileEntry(QString p, bool d, QString f) : path(p), is_dir(d), fileid(f) {}

    inline bool operator<(const FileEntry & other) const
    {
        return path < other.path;
    }
};

typedef QList<FileEntry> FileEntryList;

typedef QStringList RevisionList;

#include <QDateTime>
struct Cert
{
    enum Trust { Trusted, Untrusted } trust;
    enum Signature { Ok, Bad, Unknown } signature;
    QString key;
    QString name;
    QString value;

    QDateTime dateTimeValue() const
    {
        return QDateTime::fromString(value, "yyyy-MM-ddThh:mm:ss");
    }

    QString plainValue() const
    {
        return QString(value).replace(QRegExp("\\s+"), " ");
    }

    QString htmlValue() const
    {
        QString val(value);
        QStringList lines = val.replace("<", "&lt;").split(QRegExp("(\\r\\n|\\n|\\n\\r)"));
        QString htmlValue;
        bool inList = false;
        foreach (const QString & line, lines)
        {
            if (line.isEmpty())
            {
                if (inList)
                {
                    htmlValue += "</li>\n</ul>\n";
                    inList = false;
                    continue;
                }

                htmlValue += "<br/>\n";
                continue;
            }
            bool listStart = (0 == line.indexOf(QRegExp("^[*-]")));
            if (listStart)
            {
                if (inList)
                {
                    htmlValue += "</li>\n";
                }
                else
                {
                    inList = true;
                    htmlValue += "<ul \"margin-left: -25px\">\n";
                }

                htmlValue += "<li>" + line.mid(1).trimmed();
                continue;
            }
            if (inList)
            {
                htmlValue += " " + line.trimmed();
                continue;
            }
            htmlValue += line.trimmed() + "<br/>\n";
        }
        if (inList)
        {
            htmlValue += "</li>\n</ul>\n";
        }
        return htmlValue
            .replace(QRegExp("(\\b)([_])(\\w+)\\2"), "\\1<u>\\2\\3\\2</u>")
            .replace(QRegExp("(\\b)([*])(\\w+)\\2"), "\\1<b>\\2\\3\\2</b>")
            .replace(QRegExp("(https?://[^() ]+)"), "<a href=\"\\1\">\\1</a>")
            .replace(QRegExp("\\t"), "&nbsp;&nbsp;&nbsp;&nbsp;");
    }
};

struct CertList
{
    QString revision;
    QList<Cert> certs;

    void add(const Cert & c) { certs.push_back(c); }

    int size() const { return certs.size(); }

    Cert at(int pos) const { return certs.at(pos); }

    void clear()
    {
        certs.clear();
        revision = QString();
    }

    QStringList signers() const
    {
        QStringList signers;
        foreach (const Cert & cert, certs)
        {
            if (!signers.contains(cert.key))
                signers.push_back(cert.key);
        }
        return signers;
    }

    QList<Cert> findBySigner(const QString & key = QString()) const
    {
        QList<Cert> foundCerts;
        foreach (const Cert & cert, certs)
        {
            if (key.isEmpty() || key == cert.key)
                foundCerts.push_back(cert);
        }
        return foundCerts;
    }

    QList<Cert> findByName(const QString & name, const QString & key = QString()) const
    {
        QList<Cert> foundCerts;
        foreach (const Cert & cert, findBySigner(key))
        {
            if (cert.name != name)
                continue;
            foundCerts.push_back(cert);
        }
        return foundCerts;
    }

    QStringList findByNameValues(const QString & name, const QString & key = QString()) const
    {
        QStringList values;
        QList<Cert> certs = findByName(name, key);
        foreach (const Cert & cert, certs)
        {
            values.push_back(cert.value);
        }
        return values;
    }

    void fill(const StanzaList & stanzas)
    {
        foreach (const Stanza & st, stanzas)
        {
            Cert cert;
            foreach (const StanzaEntry & entry, st)
            {
                if (entry.sym == "key")
                {
                    I(!entry.hash.isEmpty());
                    cert.key = entry.hash;
                    continue;
                }
                if (entry.sym == "name")
                {
                    I(entry.vals.size() == 1);
                    cert.name = entry.vals.at(0);
                    continue;
                }
                if (entry.sym == "value")
                {
                    I(entry.vals.size() == 1);
                    cert.value = entry.vals.at(0).trimmed();
                    continue;
                }
                if (entry.sym == "signature")
                {
                    I(entry.vals.size() == 1);
                    QString sig = entry.vals.at(0);
                    if (sig == "ok")
                        cert.signature = Cert::Ok;
                    else if (sig == "bad")
                        cert.signature = Cert::Bad;
                    else if (sig == "unknown")
                        cert.signature = Cert::Unknown;
                    else
                        W(QString("Unknown cert signature state '%1'.").arg(sig));
                    continue;
                }
                if (entry.sym == "trust")
                {
                    I(entry.vals.size() == 1);
                    QString trust = entry.vals.at(0);
                    if (trust == "trusted")
                        cert.trust = Cert::Trusted;
                    else if (trust == "untrusted")
                        cert.trust = Cert::Untrusted;
                    else
                        W(QString("Unknown cert trust '%1'.").arg(trust));
                    continue;
                }
            }
            if (cert.key.isEmpty())
                continue;
            add(cert);
        }
    }
};

typedef QList<QByteArray> ByteArrayList;

struct Ticker
{
    QByteArray name;
    int total;
    int progress;
    bool complete;

    Ticker(const QByteArray & n = QByteArray())
        : name(n), total(0), progress(0), complete(false) {}

    /**
     * Updates this ticker's values with the values from another one
     * (of the same type)
     */
    void update(const Ticker & other)
    {
        if (name.isEmpty()) name = other.name;
        if (total < other.total) total = other.total;
        if (progress < other.progress) progress = other.progress;
        if (!complete && other.complete) complete = true;
    }

    QString toString(bool shortened = true) const
    {
        if (complete)
            return QObject::tr("%1: completed")
                .arg(QString(name));

        QString prog;
        if (shortened && progress > (1024 * 1024))
            prog = QString("%1M").arg((double)progress / (1024 * 1024), 0, 'f', 1);
        else if (shortened && progress > 1024)
            prog = QString("%1K").arg((double)progress / 1024, 0, 'f', 1);
        else
            prog = QString::number(progress);

        if (total == 0)
            return QObject::tr("%1: %2")
                .arg(QString(name)).arg(prog);

        QString tot;
        if (shortened && total > (1024 * 1024))
            tot = QString("%1M").arg((double)total / (1024 * 1024), 0, 'f', 1);
        else if (shortened && total > 1024)
            tot = QString("%1K").arg((double)total / 1024, 0, 'f', 1);
        else
            tot = QString::number(total);

        return QObject::tr("%1: %2 / %3")
            .arg(QString(name)).arg(prog).arg(tot);
    }
};

typedef QMap<char, Ticker> TickerMap;

typedef QPair<char, QByteArray> MessagePair;

#include <QSharedPointer>
class MonotoneThread;
class MonotoneTask;
typedef QSharedPointer<MonotoneThread> MonotoneThreadPtr;
typedef QSharedPointer<MonotoneTask>   MonotoneTaskPtr;

// FIXME: just a few stupid typedefs which should be expanded later on
// to more "intelligent" objects
typedef QString WorkspacePath;
typedef QString DatabaseFile;

#include <exception>
class GuitoneException : public std::exception
{
   QString message;
public:
   GuitoneException(const QString & msg) : message(msg) {}
   virtual ~GuitoneException() throw() {}
   const char * what() const throw() { return message.toStdString().c_str(); }
};

#endif

